"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation to read SPEXone DEM output

Copyright (c) 2019-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from pathlib import Path

import numpy as np

from .lib.tmtc_def import tmtc_def

# - global parameters ------------------------------


# - local functions --------------------------------
def det_dtype():
    """
    Returns numpy dtype with the registers of the SPEXone CMV4000 detector
    """
    return np.dtype([
        ('UNUSED_000', 'u1'),
        ('NUMBER_LINES', 'u1', (2)),
        ('START1', 'u1', (2)),
        ('START2', 'u1', (2)),
        ('START3', 'u1', (2)),
        ('START4', 'u1', (2)),
        ('START5', 'u1', (2)),
        ('START6', 'u1', (2)),
        ('START7', 'u1', (2)),
        ('START8', 'u1', (2)),
        ('NUMBER_LINES1', 'u1', (2)),
        ('NUMBER_LINES2', 'u1', (2)),
        ('NUMBER_LINES3', 'u1', (2)),
        ('NUMBER_LINES4', 'u1', (2)),
        ('NUMBER_LINES5', 'u1', (2)),
        ('NUMBER_LINES6', 'u1', (2)),
        ('NUMBER_LINES7', 'u1', (2)),
        ('NUMBER_LINES8', 'u1', (2)),
        ('SUB_S', 'u1', (2)),
        ('SUB_A', 'u1', (2)),
        ('MONO', 'u1'),                   # 1 bits
        ('IMAGE_FLIPPING', 'u1'),         # 2 bits
        ('INTE_SYNC', 'u1'),              # 3 bits: Int_sync, Exp_dual, Exp_ext
        ('EXP_TIME', 'u1', (3)),
        ('EXP_STEP', 'u1', (3)),
        ('EXP_KP1', 'u1', (3)),
        ('EXP_KP2', 'u1', (3)),
        ('NR_SLOPES', 'u1'),              # 2 bits
        ('EXP_SEQ', 'u1'),
        ('EXP_TIME2', 'u1', (3)),
        ('EXP_STEP2', 'u1', (3)),
        ('UNUSED_062', 'u1'),
        ('UNUSED_063', 'u1'),
        ('UNUSED_064', 'u1'),
        ('UNUSED_065', 'u1'),
        ('UNUSED_066', 'u1'),
        ('UNUSED_067', 'u1'),
        ('UNUSED_068', 'u1'),
        ('EXP2_SEQ', 'u1'),
        ('NUMBER_FRAMES', 'u1', (2)),
        ('OUTPUT_MODE', 'u1'),            # 2 bits
        ('FOT_LENGTH', 'u1'),
        ('I_LVDS_REC', 'u1'),             # 4 bits
        ('UNUSED_075', 'u1'),
        ('UNUSED_076', 'u1'),
        ('COL_CALIB', 'u1'),              # 2 bits: Col_calib, ADC_calib
        ('TRAINING_PATTERN', 'u1', (2)),  # 12 bits
        ('CHANNEL_EN', 'u1', (3)),        # 19 bits
        ('I_LVDS', 'u1'),                 # 4 bits
        ('I_COL', 'u1'),                  # 4 bits
        ('I_COL_PRECH', 'u1'),            # 4 bits
        ('I_ADC', 'u1'),                  # 4 bits
        ('I_AMP', 'u1'),                  # 4 bits
        ('VTF_L1', 'u1'),                 # 7 bits
        ('VLOW2', 'u1'),                  # 7 bits
        ('VLOW3', 'u1'),                  # 7 bits
        ('VRES_LOW', 'u1'),               # 7 bits
        ('UNUSED_092', 'u1'),
        ('UNUSED_093', 'u1'),
        ('V_PRECH', 'u1'),                # 7 bits
        ('V_REF', 'u1'),                  # 7 bits
        ('UNUSED_096', 'u1'),
        ('UNUSED_097', 'u1'),
        ('VRAMP1', 'u1'),                 # 7 bits
        ('VRAMP2', 'u1'),                 # 7 bits
        ('OFFSET', 'u1', (2)),            # 14 bits
        ('PGA_GAIN', 'u1'),               # 2 bits
        ('ADC_GAIN', 'u1'),
        ('UNUSED_104', 'u1'),
        ('UNUSED_105', 'u1'),
        ('UNUSED_106', 'u1'),
        ('UNUSED_107', 'u1'),
        ('T_DIG1', 'u1'),                 # 4 bits
        ('T_DIG2', 'u1'),                 # 4 bits
        ('UNUSED_110', 'u1'),
        ('BIT_MODE', 'u1'),               # 1 bits
        ('ADC_RESOLUTION', 'u1'),         # 2 bits
        ('PLL_ENABLE', 'u1'),             # 1 bits
        ('PLL_IN_FRE', 'u1'),             # 2 bits
        ('PLL_BYPASS', 'u1'),             # 1 bits
        ('PLL_RANGE', 'u1'),      # 8 bits: PLL range(1), out_fre(3), div(4)
        ('PLL_LOAD', 'u1'),
        ('DUMMY', 'u1'),
        ('UNUSED_119', 'u1'),
        ('UNUSED_120', 'u1'),
        ('BLACK_COL_EN', 'u1'),           # 2 bits: Black_col_en, PGA_gain
        ('UNUSED_122', 'u1'),
        ('V_BLACKSUN', 'u1'),             # 6 bits
        ('UNUSED_124', 'u1'),
        ('UNUSED_125', 'u1'),
        ('TEMP', 'u1', (2))
    ])


# - class DEMio -------------------------
class DEMio:
    """
    This class can be used to read SPEXone DEM output

    Methods
    -------
    hdr
       Returns DEM header as numpy compound array.
    number_lines
       Returns number of lines (rows).
    number_channels
       Returns number of LVDS channels used.
    lvds_clock
       Returns flag for LVDS clock, as 0: disable, 1: enable)
    pll_control
       Returns PLL control parameters: (pll_range, pll_out_fre, pll_div).
    exp_control
       Exposure time control parameters: (inte_sync, exp_dual, exp_ext).
    offset
       Returns digital offset including ADC offset
    pga_gain
       Returns PGA gain (Volt).
    temp_detector
       Returns detector temperature as raw  counts.
    exp_time(t_mcp=1e-7)
       Returns pixel exposure time [s].
    fot_time(t_mcp=1e-7)
       Returns frame overhead time [s].
    rot_time(t_mcp=1e-7)
       Returns image read-out time [s].
    frame_period(n_coad=1)
       Returns frame period [s].
    get_sci_hk()
       Returns Science telemetry, a subset of MPS and housekeeping parameters
    get_data(numlines=None)
       Returns data of a detector images (numpy uint16 array).

    Notes
    -----

    Examples
    --------
    >>> dem = DEMio(dem_file)
    >>> img_hk = dem.get_sci_hk()
    >>> img_data = dem.get_data()
    """
    def __init__(self, flname: str) -> None:
        """
        Parameters
        ----------
        flname :  str
           filename with header or binary data of DEM measurement
        """
        self.__hdr = None
        if flname.endswith('a.txt'):
            self.bin_file = flname.replace('a.txt', 'b.bin')
            self.hdr_file = flname
        elif flname.endswith('b.bin'):
            self.bin_file = flname
            self.hdr_file = flname.replace('b.bin', 'a.txt')
        else:
            raise RuntimeError('invalid filename: {}'.format(flname))

        if Path(self.hdr_file).is_file():
            self.__get_hdr()

    def __get_hdr(self) -> None:
        """
        Read DEM header data
        """
        self.__hdr = np.zeros((1,), dtype=det_dtype())
        with open(self.hdr_file, 'r') as fp:
            for line in fp:
                columns = line[:-1].split(',')
                if columns[0] == 'Reg':
                    continue

                # Fix possible errors in Name
                name = columns[2].replace(' [', '[')
                name = name.replace('_ ', '_').replace(' ', '_')

                value = int(columns[-1])
                indx = -1
                if columns[2].endswith(':0]') \
                   or columns[2].endswith('[0]') \
                   or columns[2].endswith('[2]'):
                    name = name.split('[')[0]
                    indx = 0
                elif columns[2].endswith(':8]'):
                    name = name.split('[')[0]
                    indx = 1
                elif columns[2].endswith(':16]'):
                    name = name.split('[')[0]
                    indx = 2
                elif name == 'Unused':
                    if columns[0] == '86':
                        continue
                    name = 'Unused_{:03d}'.format(int(columns[0]))

                key = name.upper()
                if isinstance(self.__hdr[0][key], np.ndarray):
                    self.__hdr[0][key][indx] = value
                else:
                    self.__hdr[0][key] = value

    @property
    def hdr(self):
        """
        Return DEM header as numpy compound array
        """
        if self.__hdr is None:
            return None

        return self.__hdr[0]

    @property
    def number_lines(self) -> int:
        """
        Return number of lines (rows)

        Register address: [1, 2]
        """
        return (self.hdr['NUMBER_LINES'][0]
                + (self.hdr['NUMBER_LINES'][1] << 8))

    @property
    def number_channels(self) -> int:
        """
        Return number of LVDS channels used
        """
        return 2 ** (4 - (self.hdr['OUTPUT_MODE'] & 0x3))

    @property
    def lvds_clock(self) -> bool:
        """
        Return flag for LVDS clock (0: disable, 1: enable)

        Register address: 82
        """
        return ((self.hdr['PLL_ENABLE'] & 0x3) == 0
                and (self.hdr['PLL_BYPASS'] & 0x3) != 0
                and (self.hdr['CHANNEL_EN'][2] & 0x4) != 0)

    def pll_control(self) -> tuple:
        """
        Returns PLL control parameters: pll_range, pll_out_fre, pll_div

        PLL_range:    range (0 or 1)
        PLL_out_fre:  output frequency (0, 1, 2 or 5)
        PLL_div:      9 (10 bit) or 11 (12 bit)

        Register address: 116
        """
        pll_div = self.hdr['PLL_RANGE'] & 0xF             # bit [0:4]
        pll_out_fre = (self.hdr['PLL_RANGE'] >> 4) & 0x7  # bit [4:7]
        pll_range = (self.hdr['PLL_RANGE'] >> 7)          # bit [7]

        return (pll_range, pll_out_fre, pll_div)

    @property
    def exp_control(self) -> tuple:
        """
        Exposure time control parameters: (inte_sync, exp_dual, exp_ext)

        Register address: 41
        """
        inte_sync = (self.hdr['INTE_SYNC'] >> 2) & 0x1
        exp_dual = (self.hdr['INTE_SYNC'] >> 1) & 0x1
        exp_ext = self.hdr['INTE_SYNC'] & 0x1

        return (inte_sync, exp_dual, exp_ext)

    @property
    def offset(self) -> int:
        """
        Returns digital offset including ADC offset

        Register address: [100, 101]
        """
        val = ((self.hdr['OFFSET'][1] << 8)
               + self.hdr['OFFSET'][0])

        return 70 + (val if val < 8192 else val - 16384)

    @property
    def pga_gain(self) -> float:
        """
        Returns PGA gain (Volt)

        Register address: 102
        """
        reg_pgagain = self.hdr['PGA_GAIN']
        # need first bit of address 121
        reg_pgagainfactor = self.hdr['BLACK_COL_EN'] & 0x1

        return (1 + 0.2 * reg_pgagain) * 2 ** reg_pgagainfactor

    @property
    def temp_detector(self) -> int:
        """
        Returns detector temperature as raw  counts

        Notes
        -----
        Uncalibrated conversion: ((1184 - 1066) * 0.3 * 40 / 40Mhz) + offs [K]
        """
        return (self.hdr['TEMP'][1] << 8) + self.hdr['TEMP'][0]

    def exp_time(self, t_mcp=1e-7):
        """
        Returns pixel exposure time [s].
        """
        # Nominal fot_length = 20, except for very short exposure_time
        reg_fot = self.hdr['FOT_LENGTH']

        reg_exptime = ((self.hdr['EXP_TIME'][2] << 16)
                       + (self.hdr['EXP_TIME'][1] << 8)
                       + self.hdr['EXP_TIME'][0])

        return 129 * t_mcp * (0.43 * reg_fot + reg_exptime)

    def fot_time(self, t_mcp=1e-7):
        """
        Returns frame overhead time [s]
        """
        # Nominal fot_length = 20, except for very short exposure_time
        reg_fot = self.hdr['FOT_LENGTH']

        return 129 * t_mcp * (reg_fot + 2 * (16 // self.number_channels))

    def rot_time(self, t_mcp=1e-7):
        """
        Returns image read-out time [s]
        """
        return 129 * t_mcp * (16 // self.number_channels) * self.number_lines

    def frame_period(self, n_coad=1):
        """
        Returns frame period [s]
        """
        return 2.38 + (n_coad
                       * (self.exp_time() + self.fot_time() + self.rot_time()))

    def get_sci_hk(self):
        """
        Returns Science telemetry, a subset of MPS and housekeeping parameters

        Returns
        -------
        numpy array
        """
        def convert_val(key):
            """
            Convert byte array to integer
            """
            val = 0
            for ii, bval in enumerate(self.__hdr[0][key]):
                val += bval << (ii * 8)

            return val

        # convert original detector parameter values to telemetry parameters
        convert_det_params = {
            'DET_NUMLINES': convert_val('NUMBER_LINES'),
            'DET_START1': convert_val('START1'),
            'DET_START2': convert_val('START2'),
            'DET_START3': convert_val('START3'),
            'DET_START4': convert_val('START4'),
            'DET_START5': convert_val('START5'),
            'DET_START6': convert_val('START6'),
            'DET_START7': convert_val('START7'),
            'DET_START8': convert_val('START8'),
            'DET_NUMLINES1': convert_val('NUMBER_LINES1'),
            'DET_NUMLINES2': convert_val('NUMBER_LINES2'),
            'DET_NUMLINES3': convert_val('NUMBER_LINES3'),
            'DET_NUMLINES4': convert_val('NUMBER_LINES4'),
            'DET_NUMLINES5': convert_val('NUMBER_LINES5'),
            'DET_NUMLINES6': convert_val('NUMBER_LINES6'),
            'DET_NUMLINES7': convert_val('NUMBER_LINES7'),
            'DET_NUMLINES8': convert_val('NUMBER_LINES8'),
            'DET_SUBS': convert_val('SUB_S'),
            'DET_SUBA': convert_val('SUB_A'),
            'DET_MONO': self.__hdr[0]['MONO'],
            'DET_IMFLIP': self.__hdr[0]['IMAGE_FLIPPING'],
            'DET_EXPCNTR': self.__hdr[0]['INTE_SYNC'],
            'DET_EXPTIME': convert_val('EXP_TIME'),
            'DET_EXPSTEP': convert_val('EXP_STEP'),
            'DET_KP1': convert_val('EXP_KP1'),
            'DET_KP2': convert_val('EXP_KP2'),
            'DET_NOFSLOPES': self.__hdr[0]['NR_SLOPES'],
            'DET_EXPSEQ': self.__hdr[0]['EXP_SEQ'],
            'DET_EXPTIME2': convert_val('EXP_TIME2'),
            'DET_EXPSTEP2': convert_val('EXP_STEP2'),
            'DET_EXP2_SEQ': self.__hdr[0]['EXP2_SEQ'],
            'DET_NOFFRAMES': convert_val('NUMBER_FRAMES'),
            'DET_OUTMODE': self.__hdr[0]['OUTPUT_MODE'],
            'DET_FOTLEN': self.__hdr[0]['FOT_LENGTH'],
            'DET_ILVDSRCVR': self.__hdr[0]['I_LVDS_REC'],
            'DET_CALIB': self.__hdr[0]['COL_CALIB'],
            'DET_TRAINPTRN': convert_val('TRAINING_PATTERN'),
            'DET_CHENA': convert_val('CHANNEL_EN'),
            'DET_ILVDS': self.__hdr[0]['I_LVDS'],
            'DET_ICOL': self.__hdr[0]['I_COL'],
            'DET_ICOLPR': self.__hdr[0]['I_COL_PRECH'],
            'DET_IADC': self.__hdr[0]['I_ADC'],
            'DET_IAMP': self.__hdr[0]['I_AMP'],
            'DET_VTFL1': self.__hdr[0]['VTF_L1'],
            'DET_VTFL2': self.__hdr[0]['VLOW2'],
            'DET_VTFL3': self.__hdr[0]['VLOW3'],
            'DET_VRSTL': self.__hdr[0]['VRES_LOW'],
            'DET_VPRECH': self.__hdr[0]['V_PRECH'],
            'DET_VREF': self.__hdr[0]['V_REF'],
            'DET_VRAMP1': self.__hdr[0]['VRAMP1'],
            'DET_VRAMP2': self.__hdr[0]['VRAMP2'],
            'DET_OFFSET': convert_val('OFFSET'),
            'DET_PGAGAIN': self.__hdr[0]['PGA_GAIN'],
            'DET_ADCGAIN': self.__hdr[0]['ADC_GAIN'],
            'DET_TDIG1': self.__hdr[0]['T_DIG1'],
            'DET_TDIG2': self.__hdr[0]['T_DIG2'],
            'DET_BITMODE': self.__hdr[0]['BIT_MODE'],
            'DET_ADCRES': self.__hdr[0]['ADC_RESOLUTION'],
            'DET_PLLENA': self.__hdr[0]['PLL_ENABLE'],
            'DET_PLLINFRE': self.__hdr[0]['PLL_IN_FRE'],
            'DET_PLLBYP': self.__hdr[0]['PLL_BYPASS'],
            'DET_PLLRATE': self.__hdr[0]['PLL_RANGE'],
            'DET_PLLLOAD': self.__hdr[0]['PLL_LOAD'],
            'DET_DETDUM': self.__hdr[0]['DUMMY'],
            'DET_BLACKCOL': self.__hdr[0]['BLACK_COL_EN'],
            'DET_VBLACKSUN': self.__hdr[0]['V_BLACKSUN'],
            'DET_T': convert_val('TEMP')
        }

        sci_hk = np.zeros((1,), dtype=np.dtype(tmtc_def(0x350)))
        for key in convert_det_params:
            sci_hk[0][key] = convert_det_params[key]

        return sci_hk

    def get_data(self, numlines=None):
        """
        Returns data of a detector frame (numpy uint16 array)

        Parameters
        ----------
        numlines : int, optional
           Provide number of detector rows when no headerfile is present
        """
        if numlines is None:
            # obtain number of rows
            numlines = self.number_lines

        # Read binary big-endian data
        return np.fromfile(self.bin_file, dtype='>u2').reshape(numlines, -1)



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Generate product name

Copyright (c) 2019-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime

import re

# - global parameters ------------------------------


# - local functions --------------------------------


# - main functions ---------------------------------
# pylint: disable=too-many-arguments
def prod_name(utc_sensing_start, msm_id=None,
              file_class='TEST', data_type='CA', level='L1A',
              orbit=None, version_number=1):
    """
    Return name of SPEXone product

    Parameters
    ----------
    utc_sensing_start: datetime.datetime
       Sensing/Validity start
    msm_id : string, optional
       Provide identifier for measurement, OCAL only
    file_class : string
       File class: OPER, CONS, TEST, ...
    data_type: string
       Product type as SPX1_<data_type>, where data type can be:
       CA (calibration data), RA (radiance data), ...
    level : string
       Level of data in product, format: ^L[012][ABCX_]$ (3 characters)
    orbit: integer, optional
       Revolution counter, zero for on-ground measurements
    version_number: integer
       Version number of the product starting at 0001

    Notes
    -----
    The general format that applies to all in-flight SPEXone products:

       PACE_CCCC_SPX1_PP_LLL_OOOOO_yyyymmddThhmmss_YYYYMMDDTHHMMSS_vvvvv.nc

    The general format that applies to all on-ground SPEXone products:

       SPX1_OCAL_<msm_id>_LLL_yyyymmddThhmmss_YYYYMMDDTHHMMSS_vvvvv.nc
    """
    # check parameters
    if re.search(r"^L[012][ABCX_]$", level) is None:
        raise ValueError("parameter: level")

    if not isinstance(version_number, int):
        raise ValueError("parameter: version_number")

    # define product ID
    if msm_id is None:
        if orbit is not None and not isinstance(orbit, int):
            raise ValueError("parameter: orbit")

        if file_class not in ('OPER', 'CONS', 'TEST'):
            raise ValueError("parameter: file_class")

        if data_type not in ('CA', 'RA'):
            raise ValueError("parameter: data_type")

        product_id = 'PACE_{:4s}_SPX1_{:2s}_{:3s}_{:05d}'.format(
            file_class, data_type, level, orbit)
    else:
        product_id = 'SPX1_OCAL_{:s}_L1A'.format(msm_id)

    # define string of sensing start as yyyymmddThhmmss
    sensing_start = utc_sensing_start.strftime("%Y%m%dT%H%M%S")

    # define instance ID
    return '{}_{:15s}_{:15s}_{:04d}.nc'.format(
        product_id, sensing_start,
        datetime.utcnow().strftime("%Y%m%dT%H%M%S"), version_number)



#!/usr/bin/env python3
"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation to convert SPEXone DEM measurements to L1A format

Copyright (c) 2019-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse

from datetime import datetime, timedelta, timezone
from pathlib import Path

import numpy as np

from pyspex import spx_product
from pyspex.lib.tmtc_def import tmtc_def
from pyspex.ccsds_io import CCSDSio
from pyspex.lv1_io import L1Aio

# - global parameters ------------------------------
LAUNCH_DATE = datetime(2022, 11, 2, tzinfo=timezone.utc)
EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)
LEAP_SECONDS = 0  # only in-flight the CCSDS packages have TAI timestamps(?)

# - local functions --------------------------------


# - main function ----------------------------------
def main():
    """
    main function
    """
    # parse command-line parameters
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description='create SPEXone Level-1A product from CCSDS packages (L0)')
    parser.add_argument('file_list', nargs='+',
                        help=("provide names of one or more files with"
                              " CCSDS packages of the same measurement"))
    parser.add_argument('--debug', action='store_true', default=False)
    parser.add_argument('--verbose', action='store_true', default=False)
    args = parser.parse_args()
    if args.verbose:
        print(args)

    args.file_list = [x for x in args.file_list if not x.endswith('.H')]

    packets = ()
    with CCSDSio(args.file_list) as ccsds:
        while True:
            packet = ccsds.read_packet()
            # print(ccsds.fp.tell(), packet is None, ccsds.packet_length)
            if packet is None or ccsds.packet_length == 0:
                break

            packets += (packet[0],)
            if args.debug:
                print('[DEBUG]: ', ccsds)

    # select NomHK packages
    nomhk_tm = ccsds.select_tm(packets, 0x320)
    # select DemHK packages
    demhk_tm = ccsds.select_tm(packets, 0x322)
    # select Science packages and combine segmented packages
    science_tm = ccsds.science_tm(packets)
    if args.debug or args.verbose:
        print('[INFO]: number of CCSDS packets ', len(packets))
        print('[INFO]: number of NomHK packages ', len(nomhk_tm))
        print('[INFO]: number of DemHK packages ', len(demhk_tm))
        print('[INFO]: number of Science images ', len(science_tm))
    del packets

    if args.debug:
        return

    # Exit because we need Science data to create a valid L1A product
    if not science_tm:
        print('[WARNING]: no science data found, exit')
        return

    # extract timestaps, telemetry and image data from Science data
    img_sec = np.empty(len(science_tm), dtype='u4')
    img_subsec = np.empty(len(science_tm), dtype='u2')
    img_id = np.empty(len(science_tm), dtype='u4')
    img_hk = np.empty(len(science_tm), dtype=np.dtype(tmtc_def(0x350)))
    img_data = []
    for ii, packet in enumerate(science_tm):
        if packet['science_hk']['ICUSWVER'] > 0x123:
            img_sec[ii] = packet['icu_time']['tai_sec']
            img_subsec[ii] = packet['icu_time']['sub_sec']
        else:
            img_sec[ii] = packet['packet_header']['tai_sec']
            img_subsec[ii] = packet['packet_header']['sub_sec']
        img_id[ii] = packet['packet_header']['sequence'] & 0x3fff
        img_hk[ii] = packet['science_hk']
        # print(packet['science_hk']['MPS_ID'],
        #       packet['science_hk']['REG_BINNING_TABLE_START'])
        img_data.append(packet['image_data'])
    img_data = np.array(img_data)

    if np.all(img_hk['ICUSWVER'] == 0x123):
        # fix bug in sub-seconds
        us100 = np.round(10000 * img_subsec.astype(float) / 65536)
        buff = us100 + img_sec - 10000
        us100 = buff.astype('u8') % 10000
        img_subsec = ((us100 << 16) // 10000).astype('u2')

    # extract timestaps and telemetry of NomHK data
    if nomhk_tm:
        nomhk_sec = np.empty(len(nomhk_tm), dtype='u4')
        nomhk_subsec = np.empty(len(nomhk_tm), dtype='u2')
        nomhk_data = np.empty(len(nomhk_tm), dtype=np.dtype(tmtc_def(0x320)))
        for ii, packet in enumerate(nomhk_tm):
            nomhk_sec[ii] = packet['packet_header']['tai_sec']
            nomhk_subsec[ii] = packet['packet_header']['sub_sec']
            nomhk_data[ii] = packet['nominal_hk']

        if np.all(img_hk['ICUSWVER'] == 0x123):
            # fix bug in sub-seconds
            us100 = np.round(10000 * nomhk_subsec.astype(float) / 65536)
            buff = us100 + nomhk_sec - 10000
            us100 = buff.astype('u8') % 10000
            med = np.median(us100)
            indx = np.where(np.abs(us100 - med) > 1000)[0]
            us100[indx] = med
            nomhk_subsec = ((us100 << 16) // 10000).astype('u2')

    # if we have DemHK data then demhk_data should be equal in size with NomHK
    if demhk_tm:
        demhk_data = np.zeros(len(nomhk_tm), dtype=np.dtype(tmtc_def(0x322)))
        for ii in range(min(len(nomhk_tm), len(demhk_tm))):
            demhk_data[ii] = demhk_tm[ii]['detector_hk']


    # generate name of L1A product
    tstamp0 = EPOCH + timedelta(seconds=int(img_sec[0]))
    if tstamp0 < LAUNCH_DATE:
        msm_id = Path(args.file_list[0]).stem.replace('_hk', '')
        try:
            new_date = datetime.strptime(
                msm_id[-22:], '%y-%j-%H:%M:%S.%f').strftime('%Y%m%dT%H%M%S.%f')
        except ValueError:
            pass
        else:
            msm_id = msm_id[:-22] + new_date

        prod_name = spx_product.prod_name(tstamp0, msm_id=msm_id)
        inflight = False
    else:
        prod_name = spx_product.prod_name(tstamp0)
        inflight = True

    # pylint: disable=unsubscriptable-object
    n_frame = 1 if img_data.ndim == 1 else img_data.shape[0]
    n_sample = img_data.size if img_data.ndim == 1 else img_data.shape[1]
    if args.verbose:
        print('[INFO]: dimension of images [{},{}]'.format(n_frame, n_sample))
    dims = {'number_of_images': n_frame,
            'samples_per_image': n_sample,
            'hk_packets': len(nomhk_tm),
            'SC_records': None}

    # Generate L1A product
    with L1Aio(prod_name, dims=dims, inflight=inflight) as l1a:
        # write image data, detector telemetry and image attributes
        l1a.fill_science(img_data, img_hk, img_id)
        l1a.fill_time(img_sec, img_subsec, group='image_attributes')

        # write engineering data
        if nomhk_tm:
            l1a.fill_nomhk(nomhk_data)
            l1a.fill_time(nomhk_sec, nomhk_subsec, group='engineering_data')

        if demhk_tm:
            l1a.fill_demhk(demhk_data)

        # write global attributes
        l1a.fill_global_attrs()
        l1a.set_attr('input_files',
                     [Path(x).name for x in args.file_list])


# --------------------------------------------------
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Add OCAL OGSE/EGSE information to a SPEXone Level-1A product

Copyright (c) 2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse

from datetime import datetime, timedelta, timezone
from pathlib import Path
import pickle

import h5py
from netCDF4 import Dataset
import numpy as np

# - global parameters ------------------------------
# enumerate source status
LDLS_DICT = {b'UNPLUGGED': 0, b'Controller Fault': 1, b'Idle': 2,
             b'Laser ON': 3, b'Lamp ON': 4, b'MISSING': 255}

# enumerate shutter positions
SHUTTER_DICT = {b'CLOSED': 0, b'OPEN': 1, b'PARTIAL': 255}


# - local functions --------------------------------
def byte_to_timestamp(str_date: str):
    """
    Helper function for numpy.loadtxt() to convert byte-string to timestamp
    """
    buff = str_date.strip().decode('ascii') + '+00:00' # date is in UTC
    return datetime.strptime(buff, '%Y%m%dT%H%M%S.%f%z').timestamp()


def read_egse(egse_file: str, verbose=False) -> tuple:
    """
    Read OGSE/EGSE data (tab separated values) to numpy compound array
    """
    # define dtype of the data
    formats = ('f8',) + 14 * ('f4',) + ('u1',) + 2 * ('i4',)\
        + ('f4', 'u1',) + 2 * ('u1',) + 3 * ('f4', 'u1',) + 7 * ('u1',)

    with open(egse_file, 'r') as fid:
        line = None
        while not line:
            line = fid.readline().strip()
            fields = line.split('\t')
            names = []
            units = []
            for field in fields:
                if field == '':
                    continue
                res = field.strip().split(' [')
                names.append(res[0].replace(' nm', 'nm').replace(' ', '_'))
                if len(res) == 2:
                    units.append(res[1].replace('[', '').replace(']', ''))
                else:
                    units.append('1')

        if 'NOMHK_packets_time' in names:
            formats = ('f8',) + formats
            convertors = {0: byte_to_timestamp,
                          1: byte_to_timestamp,
                          16: lambda s: LDLS_DICT.get(s.strip(), 255),
                          21: lambda s: SHUTTER_DICT.get(s.strip(), 255)}
        else:
            convertors = {0: byte_to_timestamp,
                          15: lambda s: LDLS_DICT.get(s.strip(), 255),
                          20: lambda s: SHUTTER_DICT.get(s.strip(), 255)}
        if verbose:
            print(len(names), names)
            print(len(formats), formats)
            print(len(units), units)

        data = np.loadtxt(fid, delimiter='\t', converters=convertors,
                          dtype={'names': names, 'formats': formats})

    return (data, units)


def write_ref_spectrum(gid):
    """
    Write reference spectrum to group /gse_data
    """
    flname = '/nfs/SPEXone/share/ckd/reference_laser_source.pickle'
    if not Path(flname).is_file():
        raise FileNotFoundError('file reference_laser_source.pickle not found')

    with open(flname, 'rb') as fp:
        data = pickle.load(fp)

    _ = gid.createDimension('wavelength', len(data['wavelength']))
    dset = gid.createVariable('wavelength', 'f8', ('wavelength',))
    dset.long_name = 'wavelength of stimulus'
    dset.units = data['units'][0]
    dset[:] = data['wavelength']

    dset = gid.createVariable('signal', 'f8', ('wavelength',))
    dset.long_name = 'signal of stimulus'
    dset.units = data['units'][1]
    dset[:] = data['signal']


def create_db_egse(db_name, egse_data, egse_units):
    """
    Write OGSE/EGSE data to HDF5 database
    """
    time_key = 'ITOS_time' if 'ITOS_time' in egse_data.dtype.names else 'time'

    with Dataset(db_name, 'w', format='NETCDF4') as fid:
        fid.creation_date = \
            datetime.now(timezone.utc).isoformat(timespec='seconds')

        _ = fid.createEnumType('u1','ldls_t',
                               {k.replace(b' ', b'_').upper(): v
                                for k, v in LDLS_DICT.items()})
        _ = fid.createEnumType('u1','shutter_t',
                               {k.upper(): v
                                for k, v in SHUTTER_DICT.items()})
        var_time = fid.createDimension('time', egse_data.size)

        var_time = fid.createVariable('time', 'f8', ('time',),
                                      chunksizes=(512,))
        indx = np.argsort(egse_data[time_key])
        var_time[:] = egse_data[time_key][indx]

        egse_t = fid.createCompoundType(egse_data.dtype, 'egse_dtype')
        var_egse = fid.createVariable('egse', egse_t, ('time',),
                                      chunksizes=(64,))
        var_egse.long_name = 'OGSE/EGSE settings'
        var_egse.units = egse_units
        var_egse.comment = ('DIG_IN_00 is of enumType ldls_t;'
                            ' SHUTTER_STATUS is of enumType shutter_t')
        var_egse[:] = egse_data[indx]


# --------------------------------------------------
def check_egse(egse_data, act_angle, alt_angle):
    """
    Check consistency of OGSE/EGSE information during measurement
    """
    for key, fmt in egse_data.dtype.fields.items():
        if fmt[0] == np.uint8:
            res_sanity = (egse_data[key] == egse_data[key][0]).all()
            if not res_sanity:
                print('[WARNING] ', key, egse_data[key])

    if act_angle:
        if not np.allclose(egse_data['ACT_ANGLE'], act_angle[0], 1e-2):
            print('[WARNING] ', 'ACT_ANGLE', egse_data['ACT_ANGLE'])
    if alt_angle:
        if not np.allclose(egse_data['ALT_ANGLE'], alt_angle[0], 1e-2):
            print('[WARNING] ', 'ALT_ANGLE', egse_data['ALT_ANGLE'])


def select_egse(l1a_file: str, egse_file: str, add_ref_laser_spectra: bool):
    """
    Write OGSE/EGSE records of a measurement to a Level-1A product
    """
    view_dict = {'M50DEG': 1, 'M20DEG': 2, '0DEG': 4, 'P20DEG': 8, 'P50DEG': 16}

    # investigate filename
    parts = l1a_file.split('_')
    vp_parts = parts[2].split('-')

    # default 0, when all viewports are illuminated
    viewport = view_dict.get(vp_parts[min(2, len(vp_parts))], 0)
    act_angle = [float(x.replace('act', ''))
                 for x in parts if x.startswith('act')]
    alt_angle = [float(x.replace('alt', ''))
                 for x in parts if x.startswith('alt')]

    # determine duration of the measurement
    with h5py.File(l1a_file, 'r') as fid:
        # pylint: disable=no-member
        msmt_start = datetime.fromisoformat(
            fid.attrs['time_coverage_start'].decode('ascii'))
        msmt_stop = datetime.fromisoformat(
            fid.attrs['time_coverage_end'].decode('ascii'))
        # print(fid.attrs['time_coverage_start'].decode('ascii'),
        #      fid.attrs['time_coverage_end'].decode('ascii'))
        duration = np.ceil((msmt_stop - msmt_start).total_seconds())

        # use the timestamp in the filename
        # pylint: disable=unsubscriptable-object
        input_file = Path(fid.attrs['input_files'][0]).stem.rstrip('_hk')
        date_str = input_file.split('_')[-1] + "+00:00"
        msmt_start = datetime.strptime(date_str, "%Y%m%dT%H%M%S.%f%z")
        msmt_start = msmt_start.replace(microsecond=0)
        msmt_stop = msmt_start + timedelta(seconds=int(duration))
        # print(msmt_start, msmt_stop)

    # open OGSE/EGSE database
    with Dataset(egse_file, 'r') as fid:
        egse_time = fid['time'][:].data
        indx = np.where((egse_time >= msmt_start.timestamp())
                        & (egse_time <= msmt_stop.timestamp()))[0]
        if indx.size == 0:
            raise RuntimeError('no OGSE/EGSE data found')

        egse_data = fid['egse'][indx[0]:indx[1]+1]
        # perform sanity check
        check_egse(egse_data, act_angle, alt_angle)

        # update Level-1A product with OGSE/EGSE information
        with Dataset(l1a_file, 'r+') as fid2:
            gid = fid2['/gse_data']
            gid['viewport'][:] = viewport

            # add OGSE/EGSE information
            _ = gid.createDimension('time', len(egse_data))
            egse_t = gid.createCompoundType(egse_data.dtype, 'egse_dtype')
            var_egse = gid.createVariable('egse', egse_t, ('time',))
            var_egse.setncatts(fid['egse'].__dict__)
            var_egse[:] = egse_data

            # write reference spectra
            if add_ref_laser_spectra:
                write_ref_spectrum(gid)

            # write sub-set of OGSE/EGSE settings as attributes
            gid.Line_skip_id = ""
            gid.Enabled_lines = np.uint16(2048)
            gid.Binning_table = ""
            gid.Binned_pixels = np.uint32(0)
            gid.Light_source = parts[2]
            gid.FOV_begin = np.nan
            gid.FOV_end = np.nan
            gid.ACT_rotationAngle = np.nan if not act_angle else act_angle[0]
            gid.ALT_rotationAngle = np.nan if not alt_angle else alt_angle[0]
            gid.ACT_illumination = np.nan
            gid.ALT_illumination = np.nan
            gid.DoLP = 0.
            gid.AoLP = 0.


# - main function ----------------------------------
def main():
    """
    Main function
    """
    # parse command-line parameters
    parser = argparse.ArgumentParser()
    parser.add_argument('--verbose', action='store_true', default=False)
    subparsers = parser.add_subparsers(help='sub-command help')
    parser_a = subparsers.add_parser('create_db',
                                     help="create new OGSE/EGSE database")
    parser_a.add_argument('--db_name', default='egse_database.nc', type=str,
                          help="name of OGSE/EGSE database")
    parser_a.add_argument('file_list', nargs='+',
                          help="provide names of one or more EGSE files (CSV)")
    parser_b = subparsers.add_parser('update',
                                     help=("add OGSE/EGSE information"
                                           " to a SPEXone Level-1A product"))
    parser_b.add_argument('--egse_db', default='egse_database.nc', type=str,
                          help="OGSE/EGSE database (HDF5)")
    parser_b.add_argument('--add_ref_laser_spectra', action='store_true',
                          default=True)
    parser_b.add_argument('l1a_file', default=None, type=str,
                          help="SPEXone L1A product")
    args = parser.parse_args()
    if args.verbose:
        print(args)

    if 'file_list' in args:
        egse = None
        for egse_file in args.file_list:
            res = read_egse(egse_file, verbose=args.verbose)
            egse = res[0] if egse is None else np.concatenate((egse, res[0]))
            units = res[1]

        create_db_egse(args.db_name, egse, units)
        return

    select_egse(args.l1a_file, args.egse_db, args.add_ref_laser_spectra)


# --------------------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation of the PACE SPEX Level-1B product

Copyright (c) 2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from netCDF4 import Dataset

# - global parameters ------------------------------


# - local functions --------------------------------


# - main function ----------------------------------
def init_l1b(l1b_flname: str, dims: dict):
    """
    Create an empty PACE SPEX Level-1B product

    Parameters
    ----------
    l1b_flname : string
       Name of Level-1B product
    dims :   dictionary
       Provide length of the Level-1B dimensions
       Default values:
            number_of_images : None     # number of image frames
            spatial_samples : 200     # sum spatial bins of all viewports
    """
    # check function parameters
    if not isinstance(dims, dict):
        raise TypeError("dims should be a dictionary")

    # initialize dimensions
    number_of_images = None
    spatial_samples = 200
    n_views = 5
    n_bins_intens = 400
    n_bins_polar = 50

    if 'number_of_images' in dims:
        number_of_images = dims['number_of_images']
    if 'samples_per_image' in dims:
        spatial_samples = dims['spatial_samples']

    # create/overwrite netCDF4 product
    rootgrp = Dataset(l1b_flname, "w")

    # create global dimensions
    _ = rootgrp.createDimension('number_of_views', n_views)
    _ = rootgrp.createDimension('spatial_samples_per_image', spatial_samples)
    _ = rootgrp.createDimension('intensity_bands_per_view', n_bins_intens)
    _ = rootgrp.createDimension('polarization_bands_per_view', n_bins_polar)
    _ = rootgrp.createDimension('bins_along_track', number_of_images)

    # create groups and all variables with attributes
    sgrp = rootgrp.createGroup('SENSOR_VIEWS_BANDS')
    dset = sgrp.createVariable("viewport_index", "u1",
                               ("spatial_samples_per_image",))
    dset.long_name = "index of viewport"
    dset.valid_min = 0
    dset.valid_max = 4
    dset.comment = "Contains indices to each viewport for all spatial samples."
    dset = sgrp.createVariable('view_angles', 'f4', ('number_of_views',))
    dset.long_name = 'along track view zenith angles at sensor'
    dset.units = 'degrees'
    dset.comment = ('view_angles is defined at the sensor, as it provides'
                    ' a swath independent value at TOA.')
    dset = sgrp.createVariable('intensity_wavelengths', 'f4',
                               ('number_of_views', 'intensity_bands_per_view'))
    dset.long_name = 'wavelength at center of intensity bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('intensity_bandpasses', 'f4',
                               ('number_of_views', 'intensity_bands_per_view'))
    dset.long_name = 'FWHM of intensity bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('polarization_wavelengths', 'f4',
                               ('number_of_views',
                                'polarization_bands_per_view'))
    dset.long_name = 'wavelength at center of polarization bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('polarization_bandpasses', 'f4',
                               ('number_of_views',
                                'polarization_bands_per_view'))
    dset.long_name = 'FWHM of polarization bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('intensity_f0', 'f4',
                               ('number_of_views', 'intensity_bands_per_view'))
    dset.long_name = 'Solar irradiance on intensity wavelength grid'
    dset.units = 'W.m-2'
    dset.comment = ('Spectral response function convolved mean solar flux'
                    ' at each intensity band and view.')
    dset = sgrp.createVariable('polarization_f0', 'f4',
                               ('number_of_views',
                                'polarization_bands_per_view'))
    dset.long_name = 'Solar irradiance on polarization wavelength grid'
    dset.units = 'W.m-2'
    dset.comment = ('Spectral response function convolved mean solar flux'
                    ' at each polarization band and view.')

    sgrp = rootgrp.createGroup('BIN_ATTRIBUTES')
    chunksizes = None if number_of_images is not None else (512,)
    dset = sgrp.createVariable('image_time', 'f8', ('bins_along_track',),
                               chunksizes=chunksizes)
    dset.long_name = 'image time (seconds of day)'
    dset.valid_min = 0
    dset.valid_max = 86400.999999
    dset.units = 'seconds'
    dset.reference = 'yyyy-mm-ddT00:00:00'

    sgrp = rootgrp.createGroup('GEOLOCATION_DATA')
    chunksizes = (None if number_of_images is not None
                  else (128, spatial_samples))
    dset = sgrp.createVariable('latitude', 'f4', ('bins_along_track',
                                                  'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.standard_name = 'latitude'
    dset.long_name = 'latitude'
    dset.valid_min = -90
    dset.valid_max = 90
    dset.units = 'degrees_north'
    dset = sgrp.createVariable('longitude', 'f4', ('bins_along_track',
                                                   'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.standard_name = 'longitude'
    dset.long_name = 'longitude'
    dset.valid_min = -180
    dset.valid_max = 180
    dset.units = 'degrees_east'
    dset = sgrp.createVariable('altitude', 'f4', ('bins_along_track',
                                                  'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.standard_name = 'altitude'
    dset.long_name = "height above mean sea level"
    dset.units = "meters"
    dset.positive = "up"
    dset.axis = "Z"
    dset = sgrp.createVariable('altitude_variability', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.long_name = 'altitude (stdev)'
    dset.units = 'meters'
    dset = sgrp.createVariable('sensor_azimuth', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.long_name = 'sensor azimuth angle'
    dset.units = 'degrees'
    dset = sgrp.createVariable('sensor_zenith', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.long_name = 'sensor zenith angle'
    dset.units = 'degrees'
    dset = sgrp.createVariable('solar_azimuth', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.long_name = 'solar azimuth angle'
    dset.units = 'degrees'
    dset = sgrp.createVariable('solar_zenith', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.long_name = 'solar zenith angle'
    dset.units = 'degrees'

    sgrp = rootgrp.createGroup('OBSERVATION_DATA')
    chunksizes = None if number_of_images is not None else \
        (8, spatial_samples, n_bins_intens)
    dset = sgrp.createVariable('I', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'intensity_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'I Stokes component'
    dset.units = 'W.m-2.sr-1.um-1'
    dset = sgrp.createVariable('I_noise', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'intensity_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of I'
    dset.units = 'W.m-2.sr-1.um-1'

    chunksizes = None if number_of_images is not None else \
        (64, spatial_samples, n_bins_polar)
    dset = sgrp.createVariable('q', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'Q over I Stokes component'
    dset.units = '1'
    dset = sgrp.createVariable('q_noise', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of Q over I'
    dset.units = '1'
    dset = sgrp.createVariable('u', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'U over I Stokes component'
    dset.units = '1'
    dset = sgrp.createVariable('u_noise', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of U over I'
    dset.units = '1'
    dset = sgrp.createVariable('AoLP', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'angle of linear polarization'
    dset.units = 'degrees'
    dset = sgrp.createVariable('AoLP_noise', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of AoLP'
    dset.units = 'degrees'
    dset = sgrp.createVariable('DoLP', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'degree of linear polarization'
    dset.units = '1'
    dset = sgrp.createVariable('DoLP_noise', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of DoLP'
    dset.units = '1'

    return rootgrp


# --------------------------------------------------
if __name__ == '__main__':
    fid = init_l1b('PACE_SPEX.20230115T123456.L1B.2.5km.V01.nc', {})
    fid.fill_global_attrs(orbit=12345, bin_size='2.5km')
    fid.close()



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation to read SPEXone instrument simulator output

Copyright (c) 2019-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from pathlib import Path

import numpy as np
try:
    import pytiff
except Exception as exc:
    raise RuntimeError('Module pytiff is required to run this module') from exc

# - global parameters ------------------------------


# - local functions --------------------------------


# - class TIFio -------------------------
class TIFio():
    """
    This class can be used to read SPEXone instrument simulator output

    Attributes
    ----------
    dir_name :  pathlib.Path
    filename :  str
    inp_tif :  bool
    lineskip :  bool

    Methods
    -------
    header()
       Read header as dictionary.
    tags()
       Return TIFF tags as dictionary.
    images(n_frame=None)
       Return TIFF data as numpy array.

    Notes
    -----

    Examples
    --------
    >>>  tif = TIFio(filename)
    >>>  print(tif.header())
    >>>  print(tif.tags()[0])
    >>>  print(tif.images().shape)
    """
    def __init__(self, hdr_file: str, inp_tif=False, lineskip=False):
        """
        Initialize class TIFio

        Parameters
        ----------
        hdr_file : str
        inp_tif : bool
        lineskip : bool

        """
        if not Path(hdr_file).is_file():
            raise FileNotFoundError('file {} not found'.format(hdr_file))

        # initialize class-attributes
        self.filename = hdr_file
        self.dir_name = Path(hdr_file).parent
        self.__stem = Path(hdr_file).stem
        self.__header = None
        self.inp_tif = inp_tif
        self.lineskip = lineskip

    def __repr__(self):
        class_name = type(self).__name__
        return '{}({!r})'.format(class_name, self.filename)

    # --------------------------------------------------
    def header(self) -> dict:
        """
        Read header as dictionary
        """
        res = {}
        with Path(self.filename).open() as fp:
            res['history'] = fp.readline()[:-1]
            _ = fp.readline()
            for line in fp:
                if line == '-\n':
                    break
                key, value = line[:-1].split(':')
                res[key.rstrip(' ')] = value.strip(' ')

            # read Spectral data stimulus
            line = fp.readline()
            buff = line[:-1].split('[')
            key = buff[0].rstrip(' ')
            if key == 'Spectral data stimulus':
                ds_sets = buff[1].replace(']', '').split(', ')

                ds_dict = {}
                for name in ds_sets:
                    line = fp.readline()
                    if not line:
                        break
                    ds_dict[name] = np.array(line[:-1].split(','), dtype=float)

                if ds_dict:
                    res[key] = ds_dict

        self.__header = res
        return res

    # --------------------------------------------------
    def tags(self) -> dict:
        """
        return TIFF tags as dictionary
        """
        if self.__header is None:
            self.header()
        n_frame = int(self.__header['Number of measurements'])

        res = []
        for num in range(n_frame):
            tif_path = self.dir_name / '{}_{}.tif'.format(self.__stem, num)
            with pytiff.Tiff(str(tif_path)) as handle:
                res.append(handle.read_tags())

        return res

    # --------------------------------------------------
    def images(self, n_frame=None):
        """
        Return TIFF data as numpy array

        Parameters
        ----------
        n_frame :  int, optional
           Distribute coadded signal (32-bit) over n_frame images (16-bit)
        """
        if self.__header is None:
            self.header()

        if self.inp_tif:
            tif_path = self.dir_name / '{}_inp.tif'.format(self.__stem)
            with pytiff.Tiff(str(tif_path)) as handle:
                data = handle[:]

            if n_frame is None:
                n_frame = 1 + (data.max() // 0xFFFF)
            elif n_frame < 1 + (data.max() // 0xFFFF):
                print('Warning: n_frame too small - precision will be lost')

            if n_frame == 1:
                return data.astype('u2')

            frames = np.zeros((n_frame,) + data.shape, dtype='u2')
            frames += (data // n_frame)
            diff = data - np.sum(frames, axis=0)
            for img in frames:
                mask = diff > 0
                img[mask] += 1
                diff[mask] -= 1

            return frames

        # convert regular TIFF files
        n_frame = int(self.__header['Number of measurements'])

        if self.lineskip:
            tif_fmt = str(self.dir_name / '{}_lineskip_{}.tif')
        else:
            tif_fmt = str(self.dir_name / '{}_{}.tif')

        res = []
        for num in range(n_frame):
            with pytiff.Tiff(tif_fmt.format(self.__stem, num)) as handle:
                res.append(handle[:])

        return np.array(res)



#!/usr/bin/env python3
"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation SPEXone instrument simulator output to L1A

Copyright (c) 2019-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse

from datetime import datetime, timezone
from pathlib import Path

import h5py
import numpy as np

from pyspex import spx_product
from pyspex.lib.tmtc_def import tmtc_def
from pyspex.tif_io import TIFio
from pyspex.lv1_io import L1Aio

# - global parameters ------------------------------
EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)


# - local functions --------------------------------
def header_as_dict(hdr, n_images):
    """
    Convert header data to Python dictionary
    """
    hdr_dict = {}
    hdr_dict['history'] = {
        'ds_type': None, 'ds_name': None, 'ds_value': None}
    hdr_dict['OCAL measurement'] = {
        'ds_type': None, 'ds_name': None, 'ds_value': None}
    hdr_dict['Row dimension'] = {
        'ds_type': None, 'ds_name': None, 'ds_value': None}
    hdr_dict['Column dimension'] = {
        'ds_type': None, 'ds_name': None, 'ds_value': None}
    hdr_dict['Number of measurements'] = {
        'ds_type': None, 'ds_name': None, 'ds_value': None}
    hdr_dict['Number of viewing angles'] = {
        'ds_type': None, 'ds_name': None, 'ds_value': None}
    hdr_dict['Integration time (s)'] = {
        'ds_type': None, 'ds_name': None, 'ds_value': None}
    hdr_dict['Exposure time (s)'] = {
        'ds_type': None, 'ds_name': None, 'ds_value': None}
    hdr_dict['Co-additions'] = {
        'ds_type': None, 'ds_name': None, 'ds_value': None}

    # keywords for binning (write as attribute
    hdr_dict['Line_skip_id'] = {
        'ds_type': 'attr',
        'ds_name': 'Line_skip_id',
        'ds_value': ''}
    if 'Line-enable-array' in hdr and hdr['Line-enable-array'] != '':
        hdr_dict['Line_skip_id']['ds_value'] = hdr['Line-enable-array']

    hdr_dict['Enabled_lines'] = {
        'ds_type': 'attr',
        'ds_name': 'Enabled_lines',
        'ds_value': np.uint16(2048)}
    if 'Enabled lines' in hdr:
        hdr_dict['Enabled_lines']['ds_value'] = np.uint16(hdr['Enabled lines'])

    hdr_dict['Binning_table'] = {
        'ds_type': 'attr',
        'ds_name': 'Binning_table',
        'ds_value': ''}
    if 'Flexible binning table' in hdr and hdr['Flexible binning table'] != '':
        hdr_dict['Binning_table']['ds_value'] = hdr['Flexible binning table']

    hdr_dict['Binned_pixels'] = {
        'ds_type': 'attr',
        'ds_name': 'Binned_pixels',
        'ds_value': np.uint32(0)}
    if 'Total flex-binned pixels' in hdr:
        hdr_dict['Binned_pixels']['ds_value'] = \
            np.uint32(hdr['Total flex-binned pixels'])

    # Datasets
    hdr_dict['Optics temperature (K)'] = {
        'ds_type': 'dset',
        'ds_name': '/engineering_data/temp_housing',
        'ds_value': np.full(n_images,
                            float(hdr['Housing temperature (K)']))}
    hdr_dict['Detector temperature (K)'] = {
        'ds_type': 'dset',
        'ds_name': '/engineering_data/temp_detector',
        'ds_value': np.full(n_images,
                            float(hdr['Detector temperature (K)']))}
    if hdr['Illuminated viewing angle'] == 'None':
        hdr_dict['Illuminated viewing angle'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
    else:
        hdr_dict['Illuminated viewing angle'] = {
            'ds_type': 'dset',
            'ds_name': '/gse_data/viewport',
            'ds_value': np.uint8(1 << int(hdr['Illuminated viewing angle']))}

    # Attributes
    if hdr['Light source'] == 'None':
        hdr_dict['Light source'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
    else:
        hdr_dict['Light source'] = {
            'ds_type': 'attr',
            'ds_name': 'Light_source',
            'ds_value': hdr['Light source'].lstrip()}
    if hdr['Illuminated fov begin'] == 'None':
        hdr_dict['Illuminated fov begin'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
    else:
        hdr_dict['Illuminated fov begin'] = {
            'ds_type': 'attr',
            'ds_name': 'FOV_begin',
            'ds_value': float(hdr['Illuminated fov begin'])}
    if hdr['Illuminated fov end'] == 'None':
        hdr_dict['Illuminated fov end'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
    else:
        hdr_dict['Illuminated fov end'] = {
            'ds_type': 'attr',
            'ds_name': 'FOV_end',
            'ds_value': float(hdr['Illuminated fov end'])}
    if hdr['Rotation stage ACT angle'] == 'None':
        hdr_dict['Rotation stage ACT angle'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
    else:
        hdr_dict['Rotation stage ACT angle'] = {
            'ds_type': 'attr',
            'ds_name': 'ACT_rotationAngle',
            'ds_value': float(hdr['Rotation stage ACT angle'])}
    if hdr['Rotation stage ALT angle'] == 'None':
        hdr_dict['Rotation stage ALT angle'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
    else:
        hdr_dict['Rotation stage ALT angle'] = {
            'ds_type': 'attr',
            'ds_name': 'ALT_rotationAngle',
            'ds_value': float(hdr['Rotation stage ALT angle'])}
    if hdr['Illuminated field ACT'] == 'None':
        hdr_dict['Illuminated field ACT'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
    else:
        hdr_dict['Illuminated field ACT'] = {
            'ds_type': 'attr',
            'ds_name': 'ACT_illumination',
            'ds_value': float(hdr['Illuminated field ACT'])}
    if hdr['Illuminated field ALT'] == 'None':
        hdr_dict['Illuminated field ALT'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
    else:
        hdr_dict['Illuminated field ALT'] = {
            'ds_type': 'attr',
            'ds_name': 'ALT_illumination',
            'ds_value': float(hdr['Illuminated field ALT'])}
    if hdr['DoLP input'] == 'None':
        hdr_dict['DoLP input'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
    else:
        hdr_dict['DoLP input'] = {
            'ds_type': 'attr',
            'ds_name': 'DoLP',
            'ds_value': float(hdr['DoLP input'])}
    if hdr['AoLP input (deg)'] == 'None':
        hdr_dict['AoLP input (deg)'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
    else:
        hdr_dict['AoLP input (deg)'] = {
            'ds_type': 'attr',
            'ds_name': 'AoLP',
            'ds_value': float(hdr['AoLP input (deg)'])}

    if hdr['Detector illumination'] == 'None':
        hdr_dict['Detector illumination'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
        hdr_dict['Det. illumination level'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}
    else:
        if 'Det. illumination level' in hdr:
            hdr_dict['Det. illumination level'] = {
                'ds_type': 'attr',
                'ds_name': 'Illumination_level',
                'ds_value': float(hdr['Det. illumination level'])}
            hdr_dict['Detector illumination'] = {
                'ds_type': 'attr',
                'ds_name': 'Light_source',
                'ds_value': hdr['Detector illumination']}
        if 'Detect. illumination e/ms' in hdr:
            hdr_dict['Det. illumination level'] = {
                'ds_type': 'attr',
                'ds_name': 'Illumination_level',
                'ds_value': float(hdr['Detect. illumination e/ms'])}
            hdr_dict['Detector illumination'] = {
                'ds_type': 'attr',
                'ds_name': 'Light_source',
                'ds_value': hdr['Detector illumination'] + ' (e.ms-1)'}

    if 'Spectral data stimulus' in hdr:
        ds_dict = hdr['Spectral data stimulus']
        hdr_dict['Wavelength stimulus'] = {
            'ds_type': 'dset',
            'ds_name': '/gse_data/wavelength',
            'ds_value': ds_dict['Wavelength (nm)']}
        hdr_dict['Radius stimulus'] = {
            'ds_type': 'dset',
            'ds_name': '/gse_data/signal',
            'ds_value': ds_dict['Radiance (photons/(s.nm.m^2.sr)']}
    else:
        hdr_dict['Spectral data stimulus'] = {
            'ds_type': None, 'ds_name': None, 'ds_value': None}

    return hdr_dict


def get_table_id(ckd_dir, bin_table_name):
    """
    Quick en dirty implementation to obtain table_id from binning-table CKD
    """
    if not Path(ckd_dir).is_dir():
        ckd_dir = '/data/richardh/SPEXone/share/ckd'

    # only read the latest version of the binning-table CKD
    bin_tables = sorted(list(Path(ckd_dir).glob('SPX1_OCAL_L1A_TBL_*.nc')))
    if not bin_tables:
        raise FileNotFoundError('No binning table found')

    table_id = None
    with h5py.File(bin_tables[-1], 'r') as fid:
        for grp in [x for x in fid.keys() if x.startswith('Table')]:
            gid = fid[grp]
            if 'origin' not in gid.attrs:
                continue

            if gid.attrs['origin'].decode('ascii') == bin_table_name:
                table_id = int(grp.split('_')[1])
                break

    return table_id


# - main function ----------------------------------
def main():
    """
    main program to illustate the creation of a L1A calibration product
    """
    parser = argparse.ArgumentParser(
        description=('create SPEXone L1A product from instrument simulations'))
    parser.add_argument('--lineskip', default=False, action='store_true',
                        help='read line-skip data, when available')
    parser.add_argument('--output', default=None,
                        help='define output directory, default=CWD')
    inp_grp = parser.add_argument_group('inp_grp', 'process inp_tif files')
    inp_grp.add_argument('--inp_tif', default=False, action='store_true',
                         help='use TIFF file with 32-bit image')
    inp_grp.add_argument('--nframe', default=None, type=int,
                         help='specify the number of images to be generated')
    parser.add_argument('--verbose', default=False, action='store_true',
                        help='be verbose')
    parser.add_argument('ascii_file', default=None,
                        help=('provide path to SPEXone instrument simulator'
                              ' product with extension ".dat"'))
    args = parser.parse_args()
    if args.verbose:
        print(args)

    if not Path(args.ascii_file).is_file():
        raise FileNotFoundError(
            'File {} does not exist'.format(args.ascii_file))

    if not args.inp_tif and args.nframe is not None:
        raise RuntimeError('option --nframe works only with "--inp_tif"')

    tif = TIFio(args.ascii_file, inp_tif=args.inp_tif, lineskip=args.lineskip)
    hdr = tif.header()
    tags = tif.tags()[0]
    images = tif.images(n_frame=args.nframe)

    # For now, we obtain the reference time from the TIFF.tags
    msm_id = hdr['OCAL measurement']
    for key in tags:
        if key.name == 'datetime':
            str_date, str_time = tags[key.value].split(' ')
            utc_start = datetime.fromisoformat(
                str_date.replace(':', '-') + 'T' + str_time + '+00:00')
            break

    prod_name = spx_product.prod_name(utc_start, msm_id=msm_id)
    if args.inp_tif:
        prod_name = prod_name.replace('_L1A_', '_inp_L1A_')
    if args.output is not None:
        dest_dir = Path(args.output)
        if not dest_dir.is_dir():
            dest_dir.mkdir()
        prod_name = str(dest_dir / prod_name)
    if args.verbose:
        print(prod_name)

    # define dimensions for L1A product
    if args.inp_tif:
        if images.ndim == 2:
            n_images = 1
            n_samples = images.size
        else:
            n_images = images.shape[0]
            n_samples = images.shape[1] * images.shape[2]
    else:
        n_images = int(hdr['Number of measurements'])
        if args.lineskip:
            n_samples = (int(hdr['Enabled lines'])
                         * int(hdr['Column dimension']))
        else:
            n_samples = (int(hdr['Row dimension'])
                         * int(hdr['Column dimension']))

    dims = {'number_of_images': n_images,
            'samples_per_image': n_samples,
            'hk_packets': n_images,
            'viewing_angles': int(hdr['Number of viewing angles'])}
    if 'Spectral data stimulus' in hdr:
        ds_dict = hdr['Spectral data stimulus']
        dims['wavelength'] = ds_dict['Wavelength (nm)'].size

    # define binning table ID
    if n_samples == 0x400000:
        table_id = 0
    elif n_samples == 0x100000:
        table_id = 1
    else:
        if args.lineskip:
            table_id = get_table_id('/nfs/SPEXone/share/ckd',
                                    hdr['Line-enable-array'])
        else:
            table_id = get_table_id('/nfs/SPEXone/share/ckd',
                                    hdr['Flexible binning table'])

    # Compute delta_time for each frame (seconds)
    # convert timestamps to seconds per day
    offs = (utc_start - EPOCH).total_seconds()
    intg_time = hdr['Co-additions'] * float(hdr['Exposure time (s)'])
    secnds = (1 + np.arange(n_images, dtype=float)) * intg_time
    img_sec = (offs + secnds // 1).astype('u4')
    sub_sec = np.round((secnds % 1) * 2**16).astype('u2')

    # Generate L1A product
    hdr_dict = header_as_dict(hdr, n_images)
    with L1Aio(prod_name, dims=dims, inflight=False) as l1a:
        # Add image data and telemetry
        l1a.set_dset('/science_data/detector_images',
                     images.reshape(n_images, -1))
        img_hk = np.zeros(n_images, dtype=np.dtype(tmtc_def(0x350)))
        l1a.set_dset('/science_data/detector_telemetry', img_hk)
        # -- no detector_telemetry!!
        #
        # Add image attributes
        # - use default values for binning_table, digital_offset, image_ID
        l1a.set_dset('/image_attributes/exposure_time',
                     np.full(n_images, float(hdr['Exposure time (s)'])))
        l1a.set_dset('/image_attributes/nr_coadditions',
                     np.full(n_images, int(hdr['Co-additions'])))

        l1a.set_dset('/image_attributes/image_ID',
                     np.arange(n_images))
        l1a.set_dset('/image_attributes/digital_offset',
                     np.zeros(n_images))
        if args.inp_tif:
            l1a.set_dset('/image_attributes/binning_table',
                         np.full(n_images, 101))
        else:
            l1a.set_dset('/image_attributes/binning_table',
                         np.full(n_images, table_id))
        l1a.fill_time(img_sec, sub_sec)
        #
        # Add engineering data
        hk_data = np.zeros(n_images, dtype=np.dtype(tmtc_def(0x320)))
        l1a.set_dset('/engineering_data/HK_telemetry', hk_data)
        l1a.set_dset('/engineering_data/HK_tlm_time', secnds)
        for key in hdr_dict:
            if hdr_dict[key]['ds_type'] != 'dset':
                continue
            l1a.set_dset(hdr_dict[key]['ds_name'], hdr_dict[key]['ds_value'])
        #
        # Add OGSE and EGSE parameters
        for key in hdr_dict:
            if hdr_dict[key]['ds_type'] != 'attr':
                continue

            l1a.set_attr(hdr_dict[key]['ds_name'],
                         hdr_dict[key]['ds_value'],
                         ds_name='gse_data')
        # Global attributes
        l1a.fill_global_attrs()
        l1a.set_attr('history', hdr['OCAL measurement'])


# --------------------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python class to handle SPEXone binning tables

Copyright (c) 2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime
from pathlib import Path

import numpy as np
from netCDF4 import Dataset

from pyspex import version


# - global parameters ------------------------------
FILL_VALUE = 0xFFFFFFFF  # 0X7FFFFFFF


# - local functions --------------------------------
class BinningTables:
    """
    Defines class to store and obtain SPEXone binning CKD

    Attributes
    ----------

    Methods
    -------

    Notes
    -----

    Examples
    --------
    """
    def __init__(self, mode='read', ckd_dir=None):
        """
        Initialize class attributes
        """
        if ckd_dir is None:
            self.ckd_dir = Path('/nfs/SPEXone/share/ckd')
            if not self.ckd_dir.is_dir():
                self.ckd_dir = Path('/data/richardh/SPEXone/share/ckd')
        else:
            self.ckd_dir = Path(ckd_dir)
        if not self.ckd_dir.is_dir():
            raise FileNotFoundError('directory with SPEXone CKD does not exist')
        self.ckd_file = None

        if mode == 'read':
            self.search_ckd()
        elif mode == 'write':
            self.create_ckd()
        else:
            self.ckd_file = mode

    def search_ckd(self):
        """
        Search CKD file with binning tables
        """
        ckd_files = list(Path(self.ckd_dir).glob('SPX1_OCAL_L1A_TBL_*.nc'))
        if not ckd_files:
            raise FileNotFoundError('No CKD with binning tables found')

        # only read the latest version of the binning-table CKD
        self.ckd_file = sorted(ckd_files)[-1]

    def create_ckd(self, release=1):
        """
        Initialize netCDF file with binning tables
        """
        # initialize netCDF file with binning tables
        self.ckd_file = 'SPX1_OCAL_L1A_TBL_{:s}_{:04d}.nc'.format(
            datetime.utcnow().strftime("%Y%m%dT%H%M%S"), release)

        with Dataset(self.ckd_dir / self.ckd_file, 'w') as fid:
            fid.setncattr('title', 'SPEXone Level-1 binning-tables')
            fid.setncattr('processing_version', version.get())
            fid.setncattr('date_created',
                          datetime.utcnow().isoformat(timespec='seconds'))
            fid.setncattr('Convensions', 'CF-1.6')
            fid.setncattr('project', 'PACE Project')
            fid.setncattr('instrument', 'SPEXone')
            fid.setncattr('institution',
                          'SRON Netherlands Institute for Space Research')

            fid.createDimension('row', 1024)
            fid.createDimension('column', 1024)

    def add_table(self, table_id: int, lineskip_arr, binning_table):
        """
        Add new binning table definition to CKD

        Parameters
        ----------
        table_id :  int
        lineskip_arr :  ndarray
        binning_table :  ndarray
          Binning table as written to DEM
        """
        index, count = np.unique(binning_table[lineskip_arr == 1, :],
                                 return_counts=True)

        with Dataset(self.ckd_dir / self.ckd_file, 'r+') as fid:
            gid = fid.createGroup('/Table_{:03d}'.format(table_id))
            gid.enabled_lines = np.uint16(lineskip_arr.sum())
            gid.flex_binned_pixels = np.uint32(index.max()+1)

            dset = gid.createVariable('binning_table', 'u4', ('row', 'column'),
                                      fill_value=FILL_VALUE,
                                      chunksizes=(128, 128),
                                      zlib=True, complevel=1, shuffle=True)
            dset.long_name = 'binning table'
            dset.valid_min = np.uint32(0)
            dset.valid_max = np.uint32(index.max())
            dset[:] = binning_table

            dset = gid.createVariable('lineskip_arr', 'u1', ('row',),
                                      zlib=True, complevel=1, shuffle=True)
            dset.long_name = 'lineskip array'
            dset.valid_min = np.uint8(0)
            dset.valid_max = np.uint8(1)
            dset[:] = lineskip_arr

            gid.createDimension('bins', count.size)
            dset = gid.createVariable('count_table', 'u2', ('bins',),
                                      zlib=True, complevel=1, shuffle=True)
            dset.long_name = 'number of aggregated pixel readings'
            dset.valid_min = np.uint16(0)
            dset.valid_max = np.uint16(count.max())
            dset[:] = count.astype('u2')

    def unbin(self, table_id: int, img_binned):
        """
        Reverse binning of detector data
        """
        with Dataset(self.ckd_dir / self.ckd_file, 'r') as fid:
            gid = fid['/Table_{:03d}'.format(table_id)]
            binning_table = gid.variables['binning_table'][:]
            lineskip_arr = gid.variables['lineskip_arr'][:]
            count_table = gid.variables['count_table'][:]

        revert = np.full(binning_table.shape, np.nan)
        table = binning_table[lineskip_arr == 1, :].reshape(-1)
        revert[lineskip_arr == 1, :] = \
            (img_binned / count_table)[table].reshape(-1, 1024)

        return revert
        
